#
# Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import os
import time
import subprocess
import yaml
import greenlet
import eventlet
import eventlet.pools
from netq_lib.common import utils
from netq_lib.orm.redisdb.models import SpiceAmberGlobal, SpiceAmberPortLevel


class SpiceDefaultConfiguration(object):
    """
    class contains default whitelist of amBer directories
    """

    def __init__(self):
        self.default_config = {
            'global': {
                'parallelism': '20',
                'period': '300',
                'timeout': '1',
                'read-fails': '3',
                'whitelist': [
                        {'xpath': '/sys/kernel/debug/nvswitch/ber/ASIC.1/MDIR'},
                        {'xpath': '/sys/kernel/debug/nvswitch/ber/ASIC.1/MGIR'},
                        {'xpath': '/sys/kernel/debug/nvswitch/ber/ASIC.1/MSGI'},
                        {'xpath': '/sys/kernel/debug/nvswitch/ber/ASIC.1/SPAD'},
                        ]
            },
            'local': [
                {'whitelist': {
                    'period': '60',
                    'read-fails': '3',
                    'timeout': '1',
                    'xpath': '/sys/kernel/debug/nvswitch/ber/ASIC.1/local_port.1/*'
                    }
                }
            ]
        }


class _SpiceConfig(object):
    """Spice Config class.
    """
    _DEF_TIMEOUT = 5  # End the cmd if not finished within this time
    _RANDOM_START = 20
    _RANDOM_STOP = 60
    _RANDOM_STEP = 10

    def __init__(self, xpath, period, read_fails, timeout=None, is_global=False):
        self.xpath = xpath
        self.period = period
        self.read_fail_allowed = read_fails
        self.isGlobal = is_global
        self.timeout = timeout if (timeout is not None) else self._DEF_TIMEOUT
        self.path_set = set()
        self.skipped_path = dict()
        self._expand_xpath()

    def _get_list_of_files(self, dir_name):
        """
        create a list of file and subdirectories
        names in the given directory
        """

        if os.path.isfile(dir_name):
            return [dir_name]
        list_of_files = os.listdir(dir_name)
        all_files = list()
        # Iterate over all the entries
        for entry in list_of_files:
            # Create full path
            full_path = os.path.join(dir_name, entry)
            # If entry is a directory then get the list of files in this directory
            if os.path.isdir(full_path):
                all_files = all_files + self._get_list_of_files(full_path)
            else:
                all_files.append(full_path)
        return all_files

    def _expand_xpath(self):
        if "*" in self.xpath:
            self.xpath = self.xpath.replace("*", "")
        if not os.path.exists(self.xpath):
            return
        path = self._get_list_of_files(self.xpath)
        self.path_set = set(path)

    def mark_path_skipped(self, path):
        count = self.skipped_path.setdefault(path, 0)
        self.skipped_path[path] = count+1

    def is_path_skipped(self, path):
        if path in self.skipped_path.keys():
            return True
        return False

    def exclude_local(self, local_path_obj):
        self.path_set = self.path_set - local_path_obj.path_set


class SpiceCollectionAgent(object):
    """
    Main class for Spice collection Framework which
    collects and stream data periodically
    """
    _BER_ROOT = '/sys/kernel/debug/nvswitch/ber/'
    _CONFIG_FILE = '/etc/netq/config.d/netq-spice-config.yml'

    def __init__(self, logger, platform):
        self._logger = logger
        self._platform = platform
        self._is_spice_loaded = False
        self._config_dict = {}
        self._pool = eventlet.GreenPool()
        self._timers = {}
        self._spice_path_objs = []
        self._global_paths = []
        self._local_paths = []
        self._default_config_obj = SpiceDefaultConfiguration()

    def __load_configuration(self):
        if os.path.exists(self._CONFIG_FILE):
            self._logger.debug("spice: loading configuration from Yaml File")
            with open(self._CONFIG_FILE, 'r') as f:
                self._config_dict = yaml.safe_load(f)
        else:
            self._logger.debug("spice: loading default config")
            self._config_dict = self._default_config_obj.default_config

    def __load_spice_module(self):
        if os.path.exists(self._BER_ROOT):
            return True
        load_cmd = "/sbin/modprobe sx_spice_dump"
        try:
            utils.subprocess_wrapper(load_cmd.split(' '))
            time.sleep(1)
            if os.path.exists(self._BER_ROOT):
                return True
        except subprocess.CalledProcessError as ex:
            print(ex)

        return False

    @staticmethod
    def __is_path_port_level(path):
        if 'local_port' in path:
            return True
        return False

    def __extract_port_from_path(self, path):
        port_num = 0
        if not self.__is_path_port_level(path):
            return port_num
        path_parents = path.split("/")
        try:
            for p in path_parents:
                if 'local_port' in p:
                    port_num = int(p.split(".")[1])
        except Exception:
            port_num = 0

        return port_num

    def _resize_green_pool(self):
        parallel_runs = int(self._config_dict['global']['parallelism'])
        self._pool.resize(parallel_runs)

    def _setup_path_whitelist(self):
        _global_path_list = self._config_dict['global']['whitelist']
        _global_period = int(self._config_dict['global']['period'])
        _global_timeout = int(self._config_dict['global']['timeout'])
        _global_read_fails = int(self._config_dict['global']['read-fails'])

        if not isinstance(_global_path_list, list):
            _global_path_list = [_global_path_list]

        for path_dict in _global_path_list:
            path = path_dict['xpath']
            if "*" in path:
                path = path.replace("*", "")
            if not os.path.exists(path):
                self._logger.debug("spice: {} not exist".format(path))
                continue
            spice_path_obj = _SpiceConfig(path, _global_period, _global_read_fails, _global_timeout, True)
            self._global_paths.append(spice_path_obj)
            self._spice_path_objs.append(spice_path_obj)

        for path in self._config_dict['local']:
            path_regex = path['whitelist']['xpath']
            period = int(path['whitelist']['period'])
            timeout = int(path['whitelist']['timeout'])
            read_fails = int(path['whitelist']['read-fails'])
            if "*" in path_regex:
                path_regex = path_regex.replace("*", "")
            if not os.path.exists(path_regex):
                self._logger.debug("spice: {} not exist".format(path_regex))
                continue
            spice_path_obj = _SpiceConfig(path_regex, period, read_fails, timeout)
            self._local_paths.append(spice_path_obj)
            self._spice_path_objs.append(spice_path_obj)

        for gpath in self._global_paths:
            for lpath in self._local_paths:
                gpath.exclude_local(lpath)

    def __update_path_timer(self, path):
        epoch_new = int(time.time())
        self._timers.setdefault(epoch_new + path.period, set())
        self._timers[epoch_new + path.period].add(path)

    def __config_path_timers(self):
        self._timers = {}
        epoch = int(time.time())
        self._timers.setdefault(epoch, set())
        for path in self._spice_path_objs:
            # rand_period = path.get_rand_period()
            self._timers[epoch].add(path)

    def __get_path_data(self, path):
        file_dict = {}
        try:
            with open(path, "r") as file:
                for line in file:
                    line = line.strip()
                    if len(line) == 0 or 'N/A' in line:
                        continue
                    line_list = line.split(":")
                    if len(line_list) < 2:
                        continue
                    file_dict[line_list[0].strip()] = line_list[-1].strip()
        except Exception as e:
            pass
        return file_dict

    def __process_spice_path(self, path_obj, path):
        try:
            if path_obj.is_path_skipped(path):
                return
            with eventlet.timeout.Timeout(path_obj.timeout, False):
                value = self.__get_path_data(path)
                eventlet.sleep(0)
            if len(value) == 0:
                path_obj.mark_path_skipped(path)
                return
            if self.__is_path_port_level(path):
                model = SpiceAmberPortLevel()
                port = self.__extract_port_from_path(path)
                model.port_num = port
            else:
                model = SpiceAmberGlobal()

            model.spice_path = path
            model.spice_value = value

            model.save()
        except Exception as e:
            self._logger.debug("spice: caught Exception {0} for path {1}".format(e, path))
            pass

    def _handle_spice_path_timer(self, epoch):
        new_timers, self._timers = self._timers.copy(), {}
        for key in new_timers:
            if epoch > key:
                for path_obj in new_timers[key]:
                    self._logger.debug("spice: processing path WL: {}".format(path_obj.xpath))
                    for path in path_obj.path_set:
                        self._pool.spawn_n(
                            self.__process_spice_path, path_obj, path)
                    self.__update_path_timer(path_obj)
                    eventlet.sleep(0.1)
            else:
                self._timers[key] = new_timers[key]

    def run(self):
        if not self._is_spice_loaded:
            self._is_spice_loaded = self.__load_spice_module()
            if not self._is_spice_loaded:
                self._logger.error("Not able to load Spice Module")
                return
            else:
                self._logger.info("spice: kernel module loaded")

        self.__load_configuration()
        self._resize_green_pool()

        # yield
        time.sleep(2)
        self._setup_path_whitelist()

        self.__config_path_timers()

        while True:
            try:
                epoch = time.time()
                # run
                self._handle_spice_path_timer(epoch)
                time.sleep(2)
            except Exception as e:
                self._logger.error("Spice Process caught Exception {}".format(e))
                break
        return
