#
# Copyright (c) 2017-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#



import time
import calendar
import os

from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import Services
from netq_lib.common.enums import (
    Daemons,
    RunningService,
    CentosRunningService,
    RhelRunningService,
    ServiceStatus
)
from netq_lib.common.utils import subprocess_wrapper, save_fullstate
from netq_agent.parsers.common.render import BaseRender


def process_frr_docker_service(pids, start_time=None):
    # process docker child pids
    # pids : dict containing the pids of individual frr daemons


    if not bool(pids):
        return

    mcls = Services.memmodel()
    routing_daemons = [Daemons.BGPD, Daemons.OSPFD,
                       Daemons.OSPF6D, Daemons.ZEBRA]

    for daemon in routing_daemons:
        pid = pids.get(daemon)
        if not pid:
            continue
        daemon_str = 'docker-' + daemon
        vrf = 'default'
        if pid == 'n/a':
            active = False
            enabled = False
            start_time = 0.0
        else:
            active = True
            enabled = True

        newsvc = Services()
        oldsvc = mcls.query.get(hostname=HOSTNAME, name=daemon_str, vrf=vrf)

        if oldsvc:
            newsvc.copy(oldsvc)
        else:
            newsvc.name = daemon_str
            newsvc.vrf = vrf

        newsvc.is_active = active
        newsvc.is_enabled = enabled
        newsvc.is_monitored = True
        newsvc.pid = pid
        newsvc.start_time = float(start_time)
        if oldsvc and oldsvc.is_active and not newsvc.is_active:
            newsvc.status = ServiceStatus.NA
        elif newsvc.is_active and (not oldsvc or not oldsvc.is_active):
            newsvc.status = ServiceStatus.OK
        newsvc.timestamp = time.time()
        newsvc.save()


class ServicesRender(BaseRender):
    def __init__(self, logger):
        super(ServicesRender, self).__init__()
        self._logger = logger

    def parse(self, cmd_out, fullstate):

        # process the output
        sysout = self.process_systemctl_output(cmd_out)

        if not bool(sysout):
            return ServiceStatus.FAIL


        # process entries in mem cache
        mcls = Services.memmodel()
        for oldsvc in mcls.query.filter(hostname=HOSTNAME):
            sysout = self.process_svc(oldsvc, sysout)

        if fullstate:
            save_fullstate(list(sysout.values()))
        else:
            # add remaining entries
            for svckey in sysout:
                entry = sysout[svckey]
                newsvc = Services()
                newsvc.copy(entry)
                newsvc.timestamp = time.time()
                newsvc.save()

    def process_svc(self, oldsvc, sysout):

        newsvc = Services()
        newsvc.copy(oldsvc)

        svckey = oldsvc.name + ':' + oldsvc.vrf
        if svckey not in sysout:
            # mark the entry as inactive
            newsvc.timestamp = time.time()
            newsvc.is_enabled = False
            newsvc.is_active = False
            newsvc.save()
            return sysout

        entry = sysout[svckey]

        update = False

        if entry.is_enabled and not oldsvc.is_enabled:
            update = True
            newsvc.is_enabled = True
        elif not entry.is_enabled and oldsvc.is_enabled:
            update = True
            newsvc.is_enabled = False

        if entry.is_active and not oldsvc.is_active:
            newsvc.is_active = True
            newsvc.status = ServiceStatus.OK
            update = True
        elif not entry.is_active and oldsvc.is_active:
            newsvc.is_active = False
            newsvc.status = ServiceStatus.NA
            update = True

        if entry.pid != oldsvc.pid:
            newsvc.pid = entry.pid
            update = True

        monitored_services = self.get_monitored_services()

        if newsvc.name in monitored_services and \
            not newsvc.is_monitored:
            update = 1
            newsvc.is_monitored = True

        if update:
            newsvc.timestamp = time.time()
            newsvc.save()

        # remove element from sysout
        del sysout[svckey]
        return sysout

    @staticmethod
    def add_routing_daemon(sysout, routing_svc, daemon, start_time):

        #It is desired to have the daemon info in the DB even when the daemon in
        # inactive.
        #if not routing_svc.is_active:
        #    return sysout

        newsvc = Services()

        svckey = daemon + ':' + routing_svc.vrf


        cmdstr = ('/bin/grep %s=yes /etc/%s/daemons' % (daemon, routing_svc.name))
        try:
            subprocess_wrapper(cmdstr.split())
        except Exception:  # pylint: disable=broad-except
            # dont process any further
            return sysout

        cmdstr = ('/usr/bin/vtysh -d %s -c quit' % daemon)
        try:
            subprocess_wrapper(cmdstr.split())
            active = True
            pidfile = '/var/run/%s/%s.pid' % (routing_svc.name, daemon)
            pid = open(pidfile, 'r').read().rstrip('\n')
        except Exception:  # pylint: disable=broad-except
            pid = 'n/a'
            active = False

        newsvc.name = daemon
        newsvc.vrf = routing_svc.vrf
        newsvc.is_active = active
        newsvc.is_enabled = routing_svc.is_enabled
        newsvc.pid = pid
        newsvc.is_monitored = True
        newsvc.timestamp = time.time()
        newsvc.start_time = float(start_time)

        if newsvc.is_active:
            newsvc.status = ServiceStatus.OK
        else:
            newsvc.status = ServiceStatus.NA

        sysout[svckey] = newsvc
        return sysout

    def process_systemctl_output(self, cmd_out):

        svcs = {}
        lines = cmd_out.splitlines()
        lineno = 0
        total_lines = len(lines)

        if os.path.isdir('/etc/frr'):
            routing_svc = RunningService.FRR
        else:
            routing_svc = RunningService.QUAGGA

        while lineno < total_lines:
            # Assumed format is the following for each systemctl entry:
            # MainPID=9808
            # Names=ssh.service
            # ActiveState=active
            # UnitFileState=enabled
            # ActiveEnterTimestamp=Wed 2017-05-17 05:56:46 UTC
            # <blank line>
            if lineno + 6 > total_lines:
                break
            if not lines[lineno:lineno + 6]:
                break
            if not (('MainPID' in lines[lineno]) and ('Names' in lines[lineno + 1]) and
                ('ActiveState' in lines[lineno + 2]) and 'UnitFileState' in lines[lineno + 3] and
                'ActiveEnterTimestamp' in lines[lineno + 4] and ('' == lines[lineno + 5])):
                next_increment_lineno = 0
                while next_increment_lineno < 6:
                    if '' == lines[lineno + next_increment_lineno]:
                        break
                    next_increment_lineno += 1

                lineno += (next_increment_lineno + 1)
                continue

            try:
                unit_state = lines[lineno + 3].split('=')[1].strip()
                if not unit_state:
                    # skip if unit_state is not parsed
                    lineno += 6
                    continue
                pid = lines[lineno].split('=')[1]
                if not pid or pid == '0':
                    pid = 'n/a'
                name = lines[lineno + 1].split('=')[1].split('.')[0]
                is_active = lines[lineno + 2].split('=')[1].strip() == 'active'
                is_enabled = (unit_state == 'enabled') or (unit_state == 'enabled-runtime')

                # Identify VRF its operating in
                if '@' in name:
                    name, vrf = name.split('@')
                    vrf = vrf.split('.')[0]
                    if 'tty' in vrf:
                        # ignore service lines such as serial-getty@ttyS0.service
                        lineno += 6
                        continue
                else:
                    vrf = 'default'

                start_time = 0.0
                try:
                    start_time_str = lines[lineno + 4].split('=')[1]
                    if start_time_str:
                        format_time = '%a %Y-%m-%d %H:%M:%S %Z'
                        start_time = calendar.timegm(time.strptime(start_time_str, format_time))
                except Exception as e:
                    self._logger.debug("Failed to parse svc start_time: {}".format(str(e)))
                    pass

                svc = Services()
                svc.name = name
                svc.vrf = vrf
                svc.is_active = is_active
                svc.is_enabled = is_enabled
                svc.pid = pid
                svc.start_time = float(start_time)

                svckey = name + ':' + vrf

                if routing_svc == name:
                    # process all the individual routing daemons
                    routing_daemons = [Daemons.BGPD, Daemons.OSPFD,
                               Daemons.OSPF6D, Daemons.ZEBRA]
                    for daemon in routing_daemons:
                        svcs = self.add_routing_daemon(svcs, svc, daemon, start_time)
                else:
                    svcs[svckey] = svc
            except IndexError:
                self._logger.debug("IndexError Exception raised while processing systemclt output section {}".format(lines[lineno:lineno + 6]))
                pass

            lineno += 6

        return svcs
