#
# Copyright (c) 2017-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import ujson as json
import ctypes
import subprocess
from netq_lib.common.enums import Platform
from netq_lib.common.utils import subprocess_wrapper, is_cl3, is_cl4, is_active_systemd_svc
from distutils.util import strtobool
from netq_lib.orm.redisdb.models import EcmpHash, EcmpInfo
from netq_lib.common.enums import ServiceStatus
from netq_lib.common import utils, configuration
from netq_lib.common.utils import save_fullstate
from netq_lib.common.utils import SWITCHD_SW_TABLE_PATH
from netq_lib.common.utils import get_cl_switchd_sw_tables_idx
from netq_agent.parsers.common.render import BaseRender
from netq_lib.common.enums import ConfKw
import re
import os

class CumulusEcmpHashRender(BaseRender):

    def __init__(self):
        super(CumulusEcmpHashRender, self).__init__()
        self.model = EcmpHash
        self.__logger = utils.get_logger(ConfKw.NETQ_AGENT, configuration)

    def search_field(self, field, text):
        return re.search(field + " = (.*)", text)

    def extract(self, field, text):
        commented_matched = self.search_field("\#.*?" + field, text)
        if(commented_matched):
            return ''
        matched = self.search_field(field, text)
        if(matched):
            return matched.groups()[0] if len(matched.groups()) > 0 else ''
        else:
            return ''

    def to_boolean(self, input, default = False):
        return strtobool(input) if input.strip() else default

    def to_int(self, input, default = -1):
        casted_unsigned_int = ctypes.c_uint32(int(input)).value if input.strip() else default
        return casted_unsigned_int

    def parse(self, cmd_output, svc_failed, **kwargs):
        # self.__logger.info("In CumulusEcmpHashRender svc_failed {} cmd_output {}".format(svc_failed, cmd_output))

        fullstate = kwargs.get('fullstate')
        platform = kwargs.get('platform', '')

        ecmp_hash_list = []

        if(svc_failed):
            self.__logger.error("Ecmp hash lookup on traffic.conf failed")
            return ServiceStatus.FAIL
        try:
            hash = EcmpHash()
            hash.symmetric_hash_enable = self.to_boolean(self.extract("symmetric_hash_enable", cmd_output), True)
            hash.resilient_hash_enable = self.to_boolean(self.extract("resilient_hash_enable", cmd_output))
            hash.seed = self.to_int(self.extract("ecmp_hash_seed", cmd_output)) # Default is random in the config
            hash.max_paths = self.to_int(self.extract("ecmp_max_paths", cmd_output), 0) # Default is 0 in the config
            if (hash.max_paths == 0):
                #SDK interprets Max path Config 0 as all available ports. report the max path from sx netdev instead of 0
                cmd = "ip -o link show type sx_netdev  | wc -l"
                hash.max_paths  = self.to_int(subprocess.check_output(cmd, shell=True))

            hash.resilient_hash_entries = self.to_int(self.extract("resilient_hash_entries_ecmp", cmd_output), 64) # Default is 64 in the config
            hash.ip_prot = self.to_boolean(self.extract("hash_config.ip_prot", cmd_output))
            hash.sip = self.to_boolean(self.extract("hash_config.sip", cmd_output))
            hash.dip = self.to_boolean(self.extract("hash_config.dip", cmd_output))
            hash.sport = self.to_boolean(self.extract("hash_config.sport", cmd_output))
            hash.dport = self.to_boolean(self.extract("hash_config.dport", cmd_output))
            hash.ip6_label = self.to_boolean(self.extract("hash_config.ip6_label", cmd_output))
            hash.ing_intf = self.to_boolean(self.extract("hash_config.ing_intf", cmd_output))
            hash.inner_ip_prot = self.to_boolean(self.extract("hash_config.inner_ip_prot", cmd_output))
            hash.inner_sip = self.to_boolean(self.extract("hash_config.inner_sip", cmd_output))
            hash.inner_dip = self.to_boolean(self.extract("hash_config.inner_dip", cmd_output))
            hash.inner_sport = self.to_boolean(self.extract("hash_config.inner_sport", cmd_output))
            hash.inner_dport = self.to_boolean(self.extract("hash_config.inner_dport", cmd_output))
            hash.inner_ip6_label = self.to_boolean(self.extract("hash_config.inner_ip6_label", cmd_output))
            if (fullstate is True):
                ecmp_hash_list.append(hash)
            else:
                hash.save()
        except Exception as e:
            self.__logger.error(e, exc_info=True)
            return ServiceStatus.FAIL

        if (fullstate is True):
            save_fullstate(ecmp_hash_list, EcmpHash().__class__.__name__)

        return ServiceStatus.OK


def process_ecmp_hash(cmd_output, svc_failed=False, **kwargs):
    render = CumulusEcmpHashRender()
    render.parse(cmd_output, svc_failed, **kwargs)


class CumulusEcmpInfoRender(BaseRender):
    def __init__(self):
        super(CumulusEcmpInfoRender, self).__init__()
        self.model = EcmpInfo
        self.__logger = utils.get_logger(ConfKw.NETQ_AGENT, configuration)
        self.intf_map_id = 0

    def get_rif_interface_mapping(self):

        if self.intf_map_id == 0:
            self.intf_map_id = get_cl_switchd_sw_tables_idx('Backend Interface Table')

        if self.intf_map_id == 0:
            try:
                if (is_cl3() or is_cl4()) and os.path.exists('/bin/systemctl'):
                    is_active = is_active_systemd_svc('switchd.service')
                    if not is_active:
                        self.__logger.debug('Cannot get intf map table, switchd service not running?')
                        return
            except Exception as e:
                self.__logger.debug('Check switchd status %s', str(e))
                return

            raise Exception('Cannot get intf map table')

        sw_table_file = '{}/{}'.format(SWITCHD_SW_TABLE_PATH, self.intf_map_id)
        cmd = 'cat {} | grep "ifp entry\|l3-interface-id"'.format(sw_table_file)

        if not os.path.exists(sw_table_file):
            return

        matched_interfaces = subprocess.check_output(cmd, shell=True)
        matched_interface_groups = re.findall("ifp entry\s+(.*?)\s+-.*?l3-interface-id\s+(\d+)", str(matched_interfaces))
        return { int(x[1]):x[0] for x in matched_interface_groups}

    def parse(self, cmd_output, svc_failed, **kwargs):

        fullstate = kwargs.get('fullstate')
        platform = kwargs.get('platform', '')

        if (svc_failed):
            self.__logger.error("Ecmp info lookup on mlxcmd failed")
            return ServiceStatus.FAIL

        try:
            rif_interface_mapping = self.get_rif_interface_mapping()
        except Exception as e:
            self.__logger.info("Ecmp info lookup on switchd software-tables l3-interface failed probably due to unsupported platform")
            self.__logger.error(e, exc_info=True)
            return ServiceStatus.FAIL

        # self.__logger.info("In CumulusEcmpInfoRender svc_failed {} cmd_output {} interfaces {}".format(svc_failed, cmd_output, rif_interface_mapping))

        json_out = json.loads(cmd_output)
        ecmp_info_list = []

        for group_id in json_out:
            for single_json_obj in json_out[group_id]['nh_list']:
                try:
                    ecmpObj = EcmpInfo()
                    ecmpObj.group_id = group_id
                    ecmpObj.rif = single_json_obj[0]
                    ecmpObj.weight = single_json_obj[1]
                    ecmpObj.bytes = single_json_obj[5]
                    ecmpObj.ifname = rif_interface_mapping.get(ecmpObj.rif, '')
                    if (fullstate is True):
                        ecmp_info_list.append(ecmpObj)
                    else:
                        ecmpObj.save()
                except Exception as e:
                    if(check_cl_version_onwards(major=5, minor=1) and Platform.MELLANOX in platform):
                        self.__logger.error(e, exc_info=True)
                    pass

        if (fullstate is True):
            save_fullstate(ecmp_info_list, EcmpInfo().__class__.__name__)

        return ServiceStatus.OK

def process_ecmp_info(cmd_output, svc_failed=False, **kwargs):
    render = CumulusEcmpInfoRender()
    render.parse(cmd_output, svc_failed, **kwargs)
